/**

Svar på delfråga a:

Detta beror antagligen på att "run_in_parallel" returnerade innan alla "worker"-trådar är klara. I
nuläget finns ingenting som ser till att "worker"-trådarna är klara innan "run_in_parallel"
returnerar, så då kan exempelvis värdena 1-5 beräknas av någon eller några av trådarna, sedan
returnerar "run_in_parallel" och main-programmet skriver ut alla resultat utan att det finns något
annat än siffran 0 i "output"-arrayen.

Kraschen beror antagligen på samma sak. Då "run_in_parallel" returnerar finns inte längre
"work_queue"-objektet, vilket skulle kunna resultera i att konstiga saker händer när
"worker"-trådarna försöker använda den.

Svar på delfråga b:

Se koden nedan. Tillägg är markerade med // <-- B

Svar på delfråga c:

I och med att "pos" ändras av alla "worker"-trådar kan följande hända:
* En tråd läser värdet 2 från "pos" och lagrar det i "item".
* En annan tråd läser värdet 2 från "pos" och lagrar det i sitt "item".
* Trådarna ökar sedan "pos" i någon ordning, så att "pos" nu är 4.

I det här fallet beräknar båda trådarna värdet för element nummer 2, medan ingen tråd kommer att
beräkna värdet för element nummer 3.

Svar på delfråga d:

Se koden nedan. Tillägg är markerade med // <-- D

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "wrap/atomics.h"
#include "wrap/synch.h"
#include "wrap/thread.h"

// Antal trådar som ska användas.
#define NUM_THREADS 4

// Funktion som kan parallelliseras.
typedef int (*work_function)(int);

// Datastruktur som delas mellan trådar och beskriver vad som ska göras och hur
// långt vi har kommit.
struct work_queue {
	// Vad ska göras?
	work_function fn;

	// Indata
	int *input;

	// Utdata.
	int *output;

	// Storlek på datan.
	int size;

	// Nästa element i 'data' som ska behandlas.
	int pos;

	// Räknar antalet worker-trådar som är klara.
	struct semaphore done_count; // <-- B

	// Lås för "pos".
	struct lock pos_lock; // <-- D
};

// Funktion som startas av 'run_in_parallel'. Hämtar arbete från 'work_queue'
// och bearbetar ett element i taget.
void worker_function(void *param) {
	struct work_queue *queue = param;
	work_function fn = queue->fn;

	while (true) {
		// Hitta ett element att bearbeta!
		lock_acquire(&queue->pos_lock); // <-- D
		int item = queue->pos;
		queue->pos++;
		lock_release(&queue->pos_lock); // <-- D

		// Är vi klara?
		if (item >= queue->size)
			break;

		// Utför arbetet!
		queue->output[item] = fn(queue->input[item]);
	}

	sema_up(&queue->done_count); // <-- B
}

// Kör funktionen 'fn' parallellt. Ska fungera som 'run_in_serial', men fördelar
// arbetet på ett antal olika trådar.
void run_in_parallel(work_function fn, int *input, int *output, int size) {
	struct work_queue queue;
	queue.fn = fn;
	queue.input = input;
	queue.output = output;
	queue.size = size;
	queue.pos = 0;
	sema_init(&queue.done_count, 0); // <-- B
	lock_init(&queue.lock); // <-- D

	for (int i = 0; i < NUM_THREADS; i++)
		thread_create("worker", 0, worker_function, &queue);

	for (int i = 0; i < NUM_THREADS; i++) // <-- B
		sema_down(&queue.done_count); // <-- B
}

// Visar hur 'run_in_parallel' bör bete sig.
void run_in_serial(work_function fn, int *input, int *output, int size) {
	for (int i = 0; i < size; i++)
		output[i] = fn(input[i]);
}


/**
 * Huvudprogram. Endast för att visa hur vi kan köra programmet. Vid rättning
 * kommer eventuella modifikationer härifrån och nedåt att ignoreras.
 */

// Enkel kvadrering.
int square(int x) {
	return x * x;
}

// Fibonaccisekvensen implementerat rekursivt.
int fib(int x) {
	if (x <= 1)
		return x;
	else
		return fib(x - 1) + fib(x - 2);
}

#define DATA_SIZE 40

int main() {
	int input[DATA_SIZE] = { 0 };
	int output[DATA_SIZE] = { 0 };
	for (int i = 0; i < DATA_SIZE; i++)
		input[i] = i;

	// Kör fibonacci parallellt.
	run_in_parallel(&fib, input, output, DATA_SIZE);

	// Kan även testa att köra "square" i stället.
	// run_in_parallel(&square, input, output, DATA_SIZE);

	printf("Resultat:\n");
	for (int i = 0; i < DATA_SIZE; i++)
		printf("%2d: %10d\n", input[i], output[i]);

	return 0;
}
